## 3.7 哨兵值控制的循环

在上一个求解班级平均分的例子中，参加考试的学生数量是固定的10人。如果事先并不知道有多少人参加考试，又当如何呢？解决这个问题的方案之一，是引入所谓的哨兵值，作为输入过程结束的标志。用户在完成所有考试成绩的输入后，输入一个特殊的哨兵值，表示没有更多数据需要输入了。程序据此退出录入循环，并开启班级平均分的计算。哨兵值的选择必须得当，即要足够特殊，不能与任何正常的输入数据冲突。对于非负的考试成绩而言，用-1作为哨兵值，不失为一种可取的做法。另外，还要考虑到用户可能在输入任何考试成绩前直接输入-1。这时在班级平均分计算中充当分母的成绩计数值为零。在C++语言中，任何数值除以零或对零求余的计算结果都是未定义的。对这种情况需要加以判断，并做出相应的处理，如打印一条错误提示。

### 3.7.1 实现哨兵值控制的循环

虽然用户输入的每个考试成绩都是整数，但班级平均分却未必是一个整数。C++语言提供了float、double和long double等数据类型来表示浮点数。这三种浮点数类型的值域依次增大，精度依次升高，即小数点右边的数字一个比一个多。

...\Ch03\SentinelLoop\SentinelLoop.cpp：

```c++
// SentinelLoop.cpp
//
// 借助哨兵值控制的循环求解平均值问题
//
#include <iostream>
#include <iomanip>
using namespace std;

int main() {
    // 初始化阶段
    int total{ 0 }; // 班级总分
    int gradeCounter{ 0 }; // 成绩计数器

    // 处理阶段，在哨兵值控制下的循环
    int grade{ 0 };
    while (grade != -1) { // 循环直到哨兵值出现
        cout << "输入第" << gradeCounter + 1
             << "位同学的考试成绩，输入-1退出："; // 提示用户输入
        cin >> grade; // 读取用户输入的考试成绩
        if (grade != -1) {
            total = total + grade; // 将考试成绩累加到班级总分中
            gradeCounter = gradeCounter + 1; // 成绩计数器加一
        }
    }

    // 最后阶段
    if (gradeCounter) {
        double average{ static_cast<double>(total) / gradeCounter }; // 计算班级平均分

        // 输出班级总分和平均分
        cout << "\n班级总分：" << total << "\n";
        cout << "班级平均分：" << setprecision(2) << fixed << average << "\n";
    }
    else {
        cout << "\n没有输入考试成绩\n";
    }
}
```

编译运行：

```
输入第1位同学的考试成绩：90
输入第2位同学的考试成绩：80
输入第3位同学的考试成绩：100
输入第4位同学的考试成绩：77
输入第5位同学的考试成绩：85
输入第6位同学的考试成绩：92
输入第7位同学的考试成绩：88
输入第8位同学的考试成绩：72
输入第9位同学的考试成绩：60
输入第10位同学的考试成绩：79
输入第11位同学的考试成绩：-1

班级总分：823
班级平均分：82
```

如前所述，整数除法只会得到整数结果。这里通过强制类型转换操作符static_cast，先将表示班级总分的整型变量total，转换成双精度浮点型，再参与班级平均分的计算，因此最后得到的结果也是双精度浮点型的数值，保存到double类型的变量average中。在这个程序的while循环语句中，使用了if条件语句，以保证作为哨兵值的-1，不会被当成考试成绩，统计到班级总分和成绩计数器中。在while循环语句之后，使用了if-else条件分支语句，以针对成绩计数值非零和为零两种情况区别对待。

#### 3.7.1.1 对比哨兵值控制的循环与计数器控制的循环
循环继续的条件不同：

- 哨兵值控制的循环以哨兵值是否出现作为循环是否继续的条件，循环次数不需要在程序设计阶段事先确定
- 计数器控制的循环以循环是否足够多作为循环是否继续的条件，循环次数必须在程序设计阶段先确定下来

读取数据的解释不同：

- 哨兵值控制的循环每次读取的数据可能是考试成绩，也可能是哨兵值，处理过程需要区别对待
- 计数器控制的循环每次读取的数据只能是考试成绩，不可能是其它值，无需对处理过程做区分

特殊情况的判断不同：

- 哨兵值控制的循环存在成绩计数值为零的情况，在除法计算中需要检测
- 计数值控制的循环成绩计数值永远不可能为零，在除法计算中无需检测

### 3.7.2 基本数据类型间的显式和隐式转换

如果用户至少输入了一个学生的考试成绩，下面的代码将用于计算班级平均分：

```c++
int total{ 0 }; // 班级总分
int gradeCounter{ 0 }; // 成绩计数器
...
double average{ static_cast<double>(total) / gradeCounter }; // 计算班级平均分
```

整数除法只会得到整数结果。即使将average声明为double类型，下面的代码依然会丢失商中的小数部分：

```c++
int total{ 0 }; // 班级总分
int gradeCounter{ 0 }; // 成绩计数器
...
double average{ total / gradeCounter }; // 计算班级平均分
```

#### 3.7.2.1 static_cast操作符
程序代码的第28行通过static_cast操作符，为其圆括号中的整型变量total，创建了一个在尖括号中指定的double类型的临时副本，后续实际参与除法计算的正是这个double类型的临时副本，这个过程被称为强制类型转换。在强制类型转换的过程中，作为转换源的total变量，其数据类型和值都没有发生丝毫变化。static_cast是C++语言提供的几种可用于强制类型转换的操作符中的一个。一切借助于强制类型转换操作符实现的类型转换都被称为显式类型转换。

#### 3.7.2.2 类型升级

程序代码的第28行通过显式类型转换，已然使班级平均分计算的分子部分成为一个double类型的数值，但它的分母部分依然是一个来自gradeCounter变量的int类型的数值。在执行任何算术运算时，编译器都要求所有参与运算的操作数必须使用相同的数据类型。因此，编译器会为整型变量gradeCounter创建一个double类型的临时副本，以与分子部分完全相同的数据类型参与到班级平均分的除法计算中。这种类型转换是隐式完成的，而且一定是从较低级别的类型（值域小，精度低）向较高级别的类型（值域大，精度高）转换。因此这种类型转换被称为隐式类型转换或类型升级。

#### 3.7.2.3 任意类型的强制类型转换操作符

基于static_cast的强制类型转换可用于所有基本数据类型和自定义类型：

```c++
static_cast<目标类型>(源类型数据) -> 目标类型数据
```

在static_cast关键字后面的尖括号中指明需要转换成何种类型。

static_cast是一个一元操作符，只有一个操作数。

### 3.7.3 浮点数格式化
#### 3.7.3.1 setprecision

程序代码的第32行调用setprecision函数，传入参数2，并将其返回值插入到标准输出流中。其作用是将浮点数的输出精度设置为两位小数。setprecision是一种流控制符，用于设置浮点数的格式化精度，需要提供表示小数位数的参数。为了使用这种带有参数的流控制符，需要包含\<iomanip\>头文件。

#### 3.7.3.2 fixed

程序代码的第32行将无参流控制符fixed插入到标准输出流中，意在指定浮点数被格式化为定点小数形式。

C++语言里的浮点数通常采用三种格式化形式：

- 浮点小数形式：有小数部分则打印小数点和小数，没有小数部分则只打印整数部分。例如：
  - $123.45$
  - $123$
- 定点小数形式：有小数部分则打印小数点和小数，没有小数部分也会打印小数点和尾随的零。例如：
  - $123.45$
  - $123.00$
- 科学计数法形式：用1到10之间的浮点数乘以10的整数次幂表示任意浮点数。例如：
  - $1.2345e2$，即$1.2345\times 10^2=123.45$
  - $1.23e2$，即$1.23\times 10^2=123$

包括setprecision和fixed在内的大部分流控制符都带有粘性，一旦设置，之后的流操作都会受其影响，直到再次设置不同的流控制符。

#### 3.7.3.3 浮点数的四舍五入
在使用setprecision和fixed流控制符格式化浮点数时，会按四舍五入处理精度，内存中的值不受影响。例如：

```c++
cout << setprecision(2) << fixed << 87.946 << ' ' << 67.543 << endl;
```

显示结果：

```
87.95 67.54
```

浮点数的默认格式：

- 不使用setprecision和fixed流控制符，C++默认保留4位小数
- 不使用setprecision流控制符，而只使用fixed流控制符，C++默认保留6位小数