## 5.8 生成随机数

为应用程序中引入某种形式的概率元素，以模拟现实世界的不确定性，这在很多游戏及与安全相关的程序中经常用到。

### 5.8.1 随机数生成器

C++11标准引入\<random\>头文件，取代传统C语言的rand函数，用于随机数的生成。一般认为传统C语言的rand函数并不具备良好的统计特征，且其产生的随机序列是可预测的，因而缺乏足够的安全性。C++11标准提供了一套更加安全的随机数生成器，可以生成既无法预测，又服从某种统计规律的随机数序列。

随机数生成器的两个主要属性：

- 引擎：随机数的生成算法
- 分布：随机数的范围、类型和统计特征

默认随机引擎类是default_random_engine，平均整数分布类是uniform_int_distribution，将二者结合起来即可得到给定范围内的，服从平均分布的，随机整数，如：

```c++
default_random_engine engine{};        // 默认随机引擎
uniform_int_distribution dist{1, 100}; // 平均整数分布
int random = dist(engine);             // 介于1到100之间服从平均分布的随机整数
```

关于随机数生成器可选引擎和分布的完整列表，参见：https://zh.cppreference.com/w/cpp/header/random。

### 5.8.2 掷六面骰子

下面的程序模拟掷十次六面骰子，得到十个介于1到6之间服从平均分布的随机点数。

...\Ch05\Die\Die.cpp：

```c++
// Die.cpp
//
// 模拟掷十次六面骰子
//
#include <iostream>
#include <random> // 随机数生成器
using namespace std;

int main() {
    default_random_engine engine{};       // 默认随机引擎
    uniform_int_distribution die{ 1, 6 }; // 平均整数分布

    // 掷十次六面骰子
    for (int roll{ 0 }; roll < 10; ++roll) {
        int const face = die(engine);
        cout << face << " ";
    }

    cout << endl;
}
```

编译运行：

```
5 1 6 6 1 6 6 2 4 2
```

在face变量上使用const限定是一个好的做法。任何初始化后不再改变其值的变量都应带有const限定。

### 5.8.3 统计规律

从上面程序的运行输出可以看到，每种点数的出现次数：

- 一点出现2次
- 二点出现2次
- 三点出现0次
- 四点出现1次
- 五点出现1次
- 六点出现4次

为了证实所掷点数服从平均分布，即每种点数出现的次数大致相等，拟增加投掷次数至六千万次。

...\Ch05\Uniform\Uniform.cpp：

```c++
// Uniform.cpp
//
// 模拟掷六千万次六面骰子，验证平均分布统计规律
//
#include <iostream>
#include <random>
#include <format>
using namespace std;

int main() {
    default_random_engine engine{};
    uniform_int_distribution die{ 1, 6 };

    int count1{ 0 }; // 一点出现的次数
    int count2{ 0 }; // 二点出现的次数
    int count3{ 0 }; // 三点出现的次数
    int count4{ 0 }; // 四点出现的次数
    int count5{ 0 }; // 五点出现的次数
    int count6{ 0 }; // 六点出现的次数

    // 掷六千万次六面骰子
    for (int roll{ 0 }; roll < 60'000'000; ++roll) {
        int const face = die(engine);

        switch (face) {
            case 1:
                ++count1;
                break;
            case 2:
                ++count2;
                break;
            case 3:
                ++count3;
                break;
            case 4:
                ++count4;
                break;
            case 5:
                ++count5;
                break;
            case 6:
                ++count6;
                break;
            default:
                cout << "无效点数" << endl;
                break;
        }
    }

    cout << format(
        "一点出现{:8}次\n"
        "二点出现{:8}次\n"
        "三点出现{:8}次\n"
        "四点出现{:8}次\n"
        "五点出现{:8}次\n"
        "六点出现{:8}次\n",
        count1, count2, count3, count4, count5, count6);
}
```

编译运行：

```
一点出现 9998437次
二点出现10000680次
三点出现10004202次
四点出现 9997565次
五点出现10000372次
六点出现 9998744次
```

在总计六千万次掷骰子的过程中，每种点数的出现的次数都接近于一千万次，符合$\frac{1}{6}$的命中概率。这就验证了其服从平均分布的统计规律。

程序代码的第44到46行是switch-case开关分支语句的default分支。该分支其实永远不可能执行到，因为变量face的值不会超出1到6的取值范围。为每个switch-case开关分支语句都提供一个default分支，以捕获可能发生的错误——即使是不可能发生的错误——是一个好的编程习惯。

事实上，对于这段程序所要实现的功能，使用数组会令代码更加简洁而优雅。

...\Ch05\UniformArray\UniformArray.cpp：

```c++
// UniformArray.cpp
//
// 模拟掷六千万次六面骰子，验证平均分布统计规律，用数组代替switch-case开关分支语句
//
#include <iostream>
#include <random>
#include <format>
using namespace std;

int main() {
    default_random_engine engine{};
    uniform_int_distribution die{ 1, 6 };

    int counts[6]{}; // 各点出现的次数

    for (int roll{ 0 }; roll < 60'000'000; ++roll) {
        int const face = die(engine);
        ++counts[face - 1]; // 统计各点出现的次数
    }

    cout << format("{:>4}{:>9}\n", "点数", "出现次数");
    for (int i = 0; i < sizeof(counts) / sizeof(counts[0]); ++i)
        cout << format("{:4}{:9}\n", i + 1, counts[i]);
}
```

编译运行：

```
点数  出现次数
  1  9998437
  2 10000680
  3 10004202
  4  9997565
  5 10000372
  6  9998744
```

### 5.8.4 随机种子

上面的程序，无论运行多少次，打印出的结果都是一样的，即所生成的所谓“随机数”是完全可以预测的，这其实并不符合常识和数学意义上的随机。默认随机引擎default_random_engine所生成的，实际上是源自特定算法的伪随机数。伪随机数所形成的序列由随机种子决定。不同的随机种子可产生不同的伪随机数序列。

...\Ch05\DieSeed\DieSeed.cpp：

```c++
// DieSeed.cpp
//
// 随机种子
//
#include <iostream>
#include <random>
using namespace std;

int main() {
    cout << "输入随机种子：";
    unsigned int seed{ 0 }; // 随机种子
    cin >> seed;

    default_random_engine engine{ seed }; // 用指定随机种子初始化默认随机引擎
    uniform_int_distribution die{ 1, 6 };

    for (int roll{ 0 }; roll < 10; ++roll) {
        int const face = die(engine);
        cout << face << " ";
    }

    cout << endl;
}
```

编译运行：

```
输入随机种子：67
4 5 6 6 5 4 2 6 1 5

输入随机种子：432
6 4 6 4 4 3 6 3 6 5

输入随机种子：67
4 5 6 6 5 4 2 6 1 5
```

如果不为随机引擎指定随机种子，随机引擎会使用默认种子产生伪随机数。

### 5.8.5 随机化

为了实现随机化，正确的做法是使用random_device类的对象来产生随机种子。random_device可以产生均匀分布的、非确定性的随机整数。这些整数是无法预测的。但出于性能的考虑，random_device仅用于为随机引擎提供随机种子，而不要直接用它来产生随机数本身，虽然它产生的随机数堪比真随机数。另外random_device产生的随机数，在某些平台上可能也是有确定性的。

...\Ch05\DieDevice\DieDevice.cpp：

```c++
// DieDevice.cpp
//
// 用真随机数作为伪随机数的随机种子，以产生接近真随机数的伪随机数
//
#include <iostream>
#include <random>
using namespace std;

int main() {
    random_device rd; // 真随机数生成器
    default_random_engine engine{ rd() }; // 用真随机数作为伪随机数的随机种子
    uniform_int_distribution die{ 1, 6 };

    for (int roll{ 0 }; roll < 10; ++roll) {
        int const face = die(engine);
        cout << face << " ";
    }

    cout << endl;
}
```

编译运行：

```
4 5 5 2 1 2 3 1 3 5

6 6 1 4 2 5 6 3 3 1

4 2 3 2 2 1 6 2 6 3
```

Visual C++实现的random_device可以提供非确定性的，可用于密码安全的真随机数。