## 5.10 作用域规则

一个标识符能在程序的哪部分使用，这部分就是该标识符的作用域。例如，在代码块中声明的局部变量，其作用域被限制在：

- 从声明位置开始到该代码块的末尾
- 该代码块中声明位置之后的嵌套块

```
 _____________
|    代码块    |
|     ...     |
|  声明局部变量 | \
|     ...     | |
|   _______   | |
|  | 嵌套块 |  | | 作用域
|  |_______|  | |
|     ...     | |
|_____________| / 
```

C++语言的作用域可被分为：

- 块作用域
- 函数作用域
- 类作用域
- 名字空间作用域
- 全局作用域

### 5.10.1 块作用域

在代码块中声明的标识符具有块作用域，它始于标识符被声明的位置，终于代码块的结束花括号

- 局部变量和函数的形参变量都具有块作用域
- 在任何块中都可以声明变量，并使其具有块作用域，包括但不限于：
  - if条件语句和if-else条件分支语句的if块、else if块和else块
  - while/do-while/for循环语句的循环体块
  - switch-case开关分支语句的case和default块
  - 任何用一对花括号括起的语句块

在嵌套块中，内层块可以声明与外层块同名的标识符

- 外层块中的标识符将被内层块中的同名标识符隐藏，直到内层块结束
- 在内层块中看到的永远是自己的局部变量，而非包含它的外层块中的同名变量
- 如果在内层块无意中声明了与外层块同名的标识符，而本意其实是希望继续使用外层块中的标识符，这通常会导致一个逻辑错误。因此，针对内层作用域隐藏外层作用域的情况，编译器通常会给出警告

可以通过static关键字将局部变量声明为静态的。例如：
```c++
static int count{ 1 };
```
- 静态局部变量同样具有块作用域，但与非静态局部变量不同的是，它在函数返回后其值仍被保留，即下次调用该函数时，静态局部变量中仍然保存着此次执行后的值
- 静态局部变量如果没有显式初始化，那么：
  - 基本类型的静态局部变量，将被默认初始化为相应类型的零值，如整型被初始为整数0，浮点型被初始化浮点数0.0，指针被初始化空指针，等等
  - 非基本类型的静态局部变量，其默认初始化依其具体类型而定，如string类型被初始化为空字符串

### 5.10.2 全局名字空间作用域

如果一个标识符的声明，不在任何函数或类的内部，那么它就具有全局名字空间作用域。这样的标识符可在任何函数中被访问。具有全局名字空间作用域的标识符可能是一个变量、一个函数或一个类。具有全局名字空间作用域的变量也被称为全局变量。它在所有函数和类之外被声明。在程序的整个执行期间，始终保留其值。

如果将一个变量声明为全局而非局部的，那么当一个本无需或不应访问该变量的函数，无意或恶意地修改其值时，极可能导致意想不到的后果。因此，除非是真正的全局资源，如cin和cout，应尽可能避免使用全局变量。这也是最小权限原则的一种体现。

最小权限原则对优秀的软件工程实践至关重要。它指出，程序代码应仅被赋予为完成特定目标所必需的权限和能力，而不是更多。局部变量的好处就在于，其仅在需要时可见，不需要它的地方也无法访问它。最小权限原则有助于防止代码意外或恶意修改不应被其访问的数据，从而使程序更加健壮，同时也令程序代码更易于阅读和维护。

一般而言，为变量提供所需的最小作用域即可。仅在特定函数中使用的变量，应被声明为该函数的局部变量，而非所有函数都能访问的全局变量。

### 5.10.3 作用域演示

下面的程序集中演示了全局变量、局部变量和静态局部变量的作用域问题。

...\Ch05\Scope\Scope.cpp：

```c++
// Scope.cpp
//
// 各种作用域
//
#include <iostream>
using namespace std;

void useLocal();       // 函数原型
void useStaticLocal(); // 函数原型
void useGlobal();      // 函数原型

int x{ 1 }; // 全局变量

int main() {
    cout << "main函数，全局变量x：" << x << '\n'; // 1

    int x{ 5 }; // main函数作用域中的局部变量，隐藏全局作用域中的同名变量

    cout << "main函数，局部变量x：" << x << '\n'; // 5

    // 块作用域
    {
        int x{ 7 }; // 内层作用域中的局部变量，隐藏外层作用域中的同名变量   

        cout << "main函数内层块，局部变量x：" << x << '\n'; // 7
    }

    cout << "main函数，局部变量x：" << x << '\n'; // 5

    useLocal();       // 使用局部变量
    useStaticLocal(); // 使用静态局部变量
    useGlobal();      // 使用全局变量

    useLocal();       // 局部变量被重新初始化
    useStaticLocal(); // 静态局部变量保留上次调用后的值
    useGlobal();      // 全局变量同样保留上次调用后的值

    cout << "\nmain函数，局部变量x：" << x << '\n'; // 5
}

// 使用局部变量
void useLocal() {
    int x{ 25 }; // 局部变量在每次调用函数时都会被重新初始化

    cout << "\nuseLocal函数入口，局部变量x：" << x << "\n"; // 25
    ++x;
    cout << "useLocal函数出口，局部变量x：" << x << "\n"; // 26
}

// 使用静态局部变量
void useStaticLocal() {
    static int x{ 50 }; // 静态局部变量仅在首次调用函数时被初始化，
                        // 之后每次调用均保留上次调用后的值
    cout << "\nuseStaticLocal函数入口，静态局部变量x：" << x << "\n"; // 50 51 ...
    ++x;
    cout << "useStaticLocal函数出口，静态局部变量x：" << x << "\n"; // 51 52 ...
}

// 使用全局变量
void useGlobal() {
    cout << "\nuseGlobal函数入口，全局变量x：" << x << "\n"; // 1 10 ...
    x *= 10; // 每次调用此函数均修改全局变量的值
    cout << "useGlobal函数出口，全局变量x：" << x << "\n"; // 10 100 ...
}
```

编译运行：

```
main函数，全局变量x：1
main函数，局部变量x：5
main函数内层块，局部变量x：7
main函数，局部变量x：5

useLocal函数入口，局部变量x：25
useLocal函数出口，局部变量x：26

useStaticLocal函数入口，静态局部变量x：50
useStaticLocal函数出口，静态局部变量x：51

useGlobal函数入口，全局变量x：1
useGlobal函数出口，全局变量x：10

useLocal函数入口，局部变量x：25
useLocal函数出口，局部变量x：26

useStaticLocal函数入口，静态局部变量x：51
useStaticLocal函数出口，静态局部变量x：52

useGlobal函数入口，全局变量x：10
useGlobal函数出口，全局变量x：100

main函数，局部变量x：5
```

#### 5.10.3.1 main函数

程序代码的第12行声明了一个名为x的全局变量，并将其初值指定为1。在任何声明了同名变量的函数或语句块中，全局变量x都会被隐藏。

程序代码的第15行打印了全局变量x的值，但在第19行打印的却是声明于第17行的，main函数的局部变量x的值。这就证明了全局变量x已被main函数的局部变量x隐藏。程序代码的第22到26行是由一对花括号括起的语句块。其中包含一个被初始化为7的局部变量x。从第25行的输出可以看出，它同时隐藏了全局变量x和main函数的局部变量x。语句块结束以后，其中的局部变量随之销毁，被其隐藏的main函数的局部变量x得以再度显现。程序代码的第28行打印了它的值。

为了演示其它作用域，程序定义了useLocal、useStaticLocal和useGlobal三个函数。它们都没有参数，也不返回任何值。程序代码的第30到32行依次调用了这三个函数，并在第34到36行再次调用了它们。程序在第38行最后打印了一次局部变量x，以证明其值并没有被那三个函数所修改。每个函数所引用的只是它们自己作用域或者全局作用域中的变量，并不会影响main函数作用域中的变量。

#### 5.10.3.2 useLocal函数

程序代码的第43行，useLocal函数声明了一个名为x的局部变量，并将其初值指定为25。随即打印了该变量的值，并在对其递增后再次打印其值。这种在函数中声明的局部变量，每次调用该函数都会重新初始化一次。因此在main函数中，两次调用useLocal函数，打印出的x的值都是25和26。

#### 5.10.3.3 useStaticLocal函数

程序代码的第52行，useStaticLocal函数声明了一个名为x的静态局部变量，并将其初值指定为50。随即打印了该变量的值，并在对其递增后再次打印其值。这种在函数中声明的静态局部变量，仅在首次调用该函数时获得初值，之后再次调用该函数，变量中保存的仍是上次调用该函数时遗留的值。因此在main函数中，第一次调用useStaticLocal函数，打印出的x的值是50和51，而在第二次调用该函数时，打印出来的却是51和52。第一次调用useStaticLocal函数后，x的值51被保留到对该函数的第二次调用，而第52行的变量初始化，仅在首次调用该函数时有效。

#### 5.10.3.4 useGlobal函数

useGlobal函数没有声明任何名为x的局部变量，因此它所引用的x，就是第12行声明并初始化的全局变量x。每次调用该函数，都会对其扩大10倍，并在扩大前后各打印一次其中的值。第一次调用该函数，打印出来的是1和10，第二次调用该函数，打印出来的是10和100。全局变量与静态局部变量一样，只被初始化一次，且不因函数返回而销毁，其中的值会一直保留到程序终止。