## 5.12 引用和引用型参数

在许多编程语言中，向函数传递参数都采用两种方式，传值和传引用。按值传参，函数得到的形参是调用语句中实参的拷贝，对形参的任何修改都不会影响实参的值。这就防止了在函数中意外修改实参所产生的副作用，这种副作用极大地妨碍了我们开发正确可靠的软件系统。但是，对于大参数而言，传值的缺点也很明显，将实参拷贝到形参，需要消耗大量的执行时间和内存空间。当然，还有另一种情况，在函数中修改实参的值，或为实参赋值，也许并非出于无意，而恰恰是程序编写者需要实现的特性。毕竟输出参数是除返回值以外，另一种使函数得以向调用者传递信息的有效方式。

### 5.12.1 引用型参数

向一个函数传递对某个变量的引用，即允许函数访问该变量本身，而非其副本。这种访问可以是读形式的，也可以是写形式的，即允许函数通过对该变量的引用，修改该变量的值。

引用型参数有利于性能，因其避免了数据复制的开销，但不利于安全，函数对实参的修改，未必出自调用者的本意。

引用型参数的本质，就是在函数中为实参指定别名。为了指明函数的某个参数是引用型的，需要在该参数的类型后面附加一个“&”符号。无论是函数的声明，还是定义，都遵循这样的约定。例如：

```c++
// 声明函数
void squareByRef(int& numberRef);

// 定义函数
void squareByRef(int& numberRef) {
    numberRef *= numberRef;
}
```

函数squareByRef的参数numberRef是一个int类型的引用。

调用函数时，不需要添加“&”符号，直接使用变量名即可。例如：

```c++
int b{ 4 };

// 调用函数
squareByRef(b);

cout << "b=" << b << " after squareByRef\n"; // b=16 after squareByRef
```

形参numberRef是对实参z的引用，修改numberRef的值就是在修改z的值。

### 5.12.2 传值和传引用

下面的程序对比了向函数传值和传引用的区别。

...\Ch05\Square\Square.cpp：

```c++
// Square.cpp
//
// 传值和传引用
//
#include <iostream>
using namespace std;

int squareByVal(int numberVal); // 传值的函数原型
void squareByRef(int& numberRef); // 传引用的函数原型

int main() {
    int a{ 3 }; // 传给squareByVal函数的参数
    int b{ 4 }; // 传给squareByRef函数的参数

    // 传值
    cout << "a=" << a << " before squareByVal\n";
    cout << "squareByVal return " << squareByVal(a) << endl;
    cout << "a=" << a << " after squareByVal\n\n";

    // 传引用
    cout << "b=" << b << " before squareByRef\n";
    squareByRef(b);
    cout << "b=" << b << " after squareByRef\n";
}

// 参数自乘，返回值为计算结果
int squareByVal(int numberVal) {
    return numberVal *= numberVal; // 不影响实参的值
}

// 参数自乘，结果保存在实参中
void squareByRef(int& numberRef) {
    numberRef *= numberRef; // 修改实参的值
}
```

编译运行：

```
a=3 before squareByVal
squareByVal return 9
a=3 after squareByVal

b=4 before squareByRef
b=16 after squareByRef
```

调用squareByVal和squareByRef函数采用了相同的风格，都是直接在实参表中使用变量名（a和b）。但是，编译器会检查函数的声明和定义，以确定参数的传递是按值（int numberVal）还是按引用（int& numberRef）。

### 5.12.3 引用即别名

引用也可以在除函数参数以外的场合中使用，但它仍然表示另一个变量的别名。例如：

```c++
int count{ 1 }; // 声明一个整型变量count
int& countRef{ count }; // 声明一个对整型变量的引用countRef，并使之成为变量count的别名
++countRef; // 通过count变量的别名countRef令其自增
```

引用型变量必须在声明时初始化，而且不能作为引用再成为其它变量的别名。从这个角度说，引用是常量。对引用所做的所有操作，实际都是对其目标变量做的。引用只是其目标变量的另一个名字。引用的初始化器必须是一个左值，即可以出现在赋值操作符左边的东西，而不能是常量或者右值，即只能出现在赋值操作符右边的东西。

### 5.12.4 常引用

为了限制一个函数不能对它的某个引用型参数的值做修改，可在该参数的类型中使用const限定符。

displayName函数用于在标准输出上打印从参数传入的姓名字符串：

```c++
void displayName(string name) {
    cout << name << endl;
}
```

如果以传值的方式接受参数，由于string对象可能很大，这种复制操作会严重影响程序的性能。因此，这里使用引用型参数更为合理：

```c++
void displayName(string& name) {
    cout << name << endl;
}
```

同时考虑到displayName函数并不需要修改参数的值。根据最小权限原则，将其参数声明为常引用，不失为一种明智之举：

```c++
void displayName(const string& name) {
    cout << name << endl;
}
```

name是一个对只读字符串对象的引用。凭借引用，获得了性能上的优势，同时又只读，限制了函数对其目标实参的（意外）修改。性能安全两不误。

### 5.12.5 返回对局部变量的引用

如果从函数中返回一个对非静态局部变量的引用，该引用的目标变量将在函数返回后自动被销毁，引用随即失效。访问这样的引用会导致未定义的后果，包括但不限于程序崩溃或数据破坏。

对已销毁变量的引用称为悬空引用。访问这样的引用被视为一种逻辑错误，编译器通常会为此发出警告。

软件工程团队的一般性规定是，代码在部署之前，必须在没有任何警告的前提下通过编译。编译器的某些选项，如GCC的“-Wall”，可以给出所有警告，而另一些选择，如GCC的“-Werror”，则将警告视为错误。有意识地使用这些选项，可以强制开发团队遵守上述规定。