## 5.15 函数重载

C++允许在同一个作用域中定义名字相同的函数，只要它们的签名不同即可，这种做法成为函数重载。C++编译器根据调用函数时，所传实参的个数、类型和顺序，选择匹配度最高的重载版本进行调用。

利用函数重载，可以创建一组执行相似任务，但对不同类型的数据，又有不同处理方式的同名函数。

数学库中的很多函数都有针对不同数据类型的重载版本。

对执行密切相关任务的函数进行重载，可使程序代码更加清晰。

### 5.15.1 重载的square函数

...\Ch05\OverloadedSquare\OverloadedSquare.cpp：

```c++
// OverloadedSquare.cpp
//
// 重载的square函数
//
#include <iostream>
using namespace std;

// 整型版的square函数
int square(int x) {
    cout << "整数" << x << "的平方：";
    return x * x;
}

// 双精度浮点型版的square函数
double square(double x) {
    cout << "双精度浮点数" << x << "的平方：";
    return x * x;
}

int main() {
    cout << square(7) << endl; // 调用整型版的square函数
    cout << square(7.5) << endl; // 调用双精度浮点型版的square函数
}
```

编译运行：

```
整数7的平方：49
双精度浮点数7.5的平方：56.25
```

程序代码的第9到12行和第15到18行，定义了两个名为square函数，分别用于计算整数和双精度浮点数的平方。它们的函数名虽然相同，但参数的类型不同，一个为int，另一个却为double，符合有关函数重载的定义。

程序代码的第21行，编译器根据所提供的参数——字面值7——整数，选择了整型版的square函数，完成调用；程序代码的第22行，编译器根据所提供的参数——字面值7.5——双精度浮点数，选择了双精度浮点型版的square函数，完成调用。每种情况下，编译器总会根据实参的类型，选择最适当的重载版本。

### 5.15.2 编译器如何区分不同的重载版本？

重载函数是通过其签名来区分的。函数的签名是函数名及其参数类型（按顺序）的组合。类型安全连接根据实参与形参的类型一致性，确保所调用的函数版本正确。为了实现类型安全连接，编译器会在其内部，利用函数参数的类型信息，对函数标识符进行编码，这个过程称为名称改编。名称改编的具体规则因编译器而异，因此最终提交给链接器，用于生成可执行程序的目标模块和库，必须出自相同或至少是兼容的编译器。

...\Ch05\NameMangling\NameMangling.cpp：

```c++
// NameMangling.cpp
//
// 基于名称改编的类型安全连接
//
// 整型版的square函数
int square(int x) {
   return x * x;
}

// 双精度浮点型版的square函数
double square(double x) {
   return x * x;
}

// 接收int、float、char和int&型参数，无返回值的函数
void nothing1(int a, float b, char c, int& d) {  }

// 接收char、int、float&和double&型参数，返回int型值的函数
int nothing2(char a, int b, float& c, double& d) {
   return 0;
}

int main() { }
```

编译得到目标文件：

```bash
g++ -c NameMangling.cpp
```

查看目标文件中的符号表：

```bash
nm NameMangling.o
```

```
...
0000000b T __Z6squared
00000000 T __Z6squarei
00000024 T __Z8nothing1ifcRi
00000033 T __Z8nothing2ciRfRd
...
```

GNU编译器的名称改编规则：双下划线前缀+大写字母Z+原始函数名长度+原始函数名+类型标签序列

- d：double
- i：int
- f：float
- c：char
- Ri：int&
- Rf：float&
- Rd：double&

注意，函数的返回类型并不是函数签名的一部分，也不参与名称改编。重载的函数可以有相同或不同的返回类型，但必须有不同的形参表。

调用带有缺省参数的函数，可能会与该函数的另一个重载版本构成冲突，这会引发编译错误。例如：

```c++
...
void foo(int x, int y = 2);
void foo(int x);
...
foo(1); // 歧义
...
```

### 5.15.3 重载解析详情

有关编译器解析函数重载的详细信息，参见：https://timsong-cpp.github.io/cppwp/n4861/over.match。

语言标准中有关函数重载决议的细节，参见：https://zh.cppreference.com/w/cpp/language/overload_resolution。

### 5.15.4 操作符重载

第11章会详细讨论有关操作符重载的细节。通过操作符重载，可以令许多本来用于操作基本类型数据的操作符，也能被用于操作用户自定义类型的数据。

流插入操作符“<<”和流提取操作符“>>”之所以能胜任针对不同类型数据的流插入和流提取操作，是因为C++标准库已经为这两个操作符定义了针对不同数据类型的重载版本。程序设计者还可以为其定义更多的重载版本，以支持更多用户自定义的数据类型。