## 5.16 函数模板

函数重载通常用于，针对不同类型的数据，执行相似操作的场合。而如果针对每种类型的数据，执行的操作步骤和程序逻辑完全一样，则可以考虑利用函数模板获得更加简洁、精炼的重载效果。根据调用函数模板时提供的实参类型，编译器会自动生成针对每种具体类型的重载版本，即函数模板实例，以与实参相匹配的类型完成函数调用。这样做的好处是，只要定义一个函数模板，就相当于定义了一整套重载函数。基于模板的编程称为泛型编程。

### 5.16.1 定义maximum函数模板

...\Ch05\FuncTemp\maximum.h：

```c++
// maximum.h
//
// 定义maximum函数模板
//
#pragma once

// 求三个数中最大值的函数模板
template <typename T>  // template<class T>
T maximum(T x, T y, T z) {
    T max{ x }; // 假设x是最大值    

    // 如果最大值比y小，则y是最大值
    if (max < y) {
        max = y;
    }

    // 如果最大值比z小，则z是最大值
    if (max < z) {
        max = z;
    }

    return max; // 返回最大值
}
```

这里定义了一个名为maximum的函数模板，用于求取三个数中的最大值。所有函数模板的定义都以template关键字开始，后跟一个被一对尖括号括起来的模板参数列表。模板参数列表中的每个类型参数前，都冠以typename或class关键字。类型参数是包括基本类型和用户自定义类型在内的某个具体类型的占位符。这些占位符，如本例中的T，被用于指定函数参数的类型，如第9行，指定函数返回值的类型，如第9行，指定函数中局部变量的类型，如第10行，等等。定义函数模板和定义普通函数并没有太大区别，只是使用了类型参数作为实际数据类型的占位符。

maximum函数模板的类型参数T是一个类型占位符，表示该函数要处理的数据类型。在一个特定函数模板的参数列表中，每个模板参数的名称都必须是唯一的。编译器在程序代码中看到对maximum的调用时，会先用实参的具体类型，替换掉整个模板定义中的类型占位符T，从而创建一个完整的函数模板实例，用于求取三个实参类型值中的最大值，然后再将该函数模板实例编译为二进制代码。由此不难看出，函数模板实际上只是一种指导编译器生成源代码的手段，真正被执行的都是类型明确的具体函数，而绝非函数模板本身。

### 5.16.2 使用maximum函数模板

下面的程序调用之前定义的maximum函数模板，分别求取三个整数、三个浮点数和三个字符中的最大值。

...\Ch05\FuncTemp\FuncTemp.cpp：

```c++
// FuncTemp.cpp
//
// 使用maximum函数模板
// 
#include <iostream>
using namespace std;

#include "maximum.h" // 包含定义了maximum函数模板的头文件

int main() {
    // 求取三个整数的最大值
    cout << "输入三个整数：";
    int n1, n2, n3;
    cin >> n1 >> n2 >> n3;

    // 调用int版maximum函数
    cout << "最大整数：" << maximum(n1, n2, n3) << endl;

    cout << endl;

    // 求取三个浮点数的最大值
    cout << "输入三个浮点数：";
    double d1, d2, d3;
    cin >> d1 >> d2 >> d3;

    // 调用double版maximum函数
    cout << "最大浮点数：" << maximum(d1, d2, d3) << endl;

    cout << endl;

    // 求取三个字符的最大值
    cout << "输入三个字符：";
    char ch1, ch2, ch3;
    cin >> ch1 >> ch2 >> ch3;

    // 调用char版maximum函数
    cout << "最大字符：" << maximum(ch1, ch2, ch3) << endl;
}
```

编译运行：

```
输入三个整数：1 3 2
最大整数：3

输入三个浮点数：1.1 3.3 2.2
最大浮点数：3.3

输入三个字符：A C B
最大字符：C
```

程序代码的第17、27和37行分别以三个int类型、double类型和char类型的实参，调用maximum函数模板。编译器在幕后为每个调用单独创建了三个分别针对int类型、double类型和char类型重载实例：

- maximum函数模板针对int类型的重载实例
  ```c++
  int maximum(int x, int y, int z) {
      int max{ x };    
  
      if (max < y) {
          max = y;
      }
  
      if (max < z) {
          max = z;
      }
  
      return max;
  }
  ```
- maximum函数模板针对double类型的重载实例
  ```c++
  double maximum(double x, double y, double z) {
      int max{ x };    
  
      if (max < y) {
          max = y;
      }
  
      if (max < z) {
          max = z;
      }
  
      return max;
  }
  ```
- maximum函数模板针对char类型的重载实例
  ```c++
  char maximum(char x, char y, char z) {
      int max{ x };    
  
      if (max < y) {
          max = y;
      }
  
      if (max < z) {
          max = z;
      }
  
      return max;
  }
  ```